import threading
from typing import Any, Callable

from port_ocean.consumers.kafka_consumer import KafkaConsumer, KafkaConsumerConfig
from port_ocean.context.ocean import (
    PortOceanContext,
    initialize_port_ocean_context,
    ocean,
)
from port_ocean.core.trigger_channel.base import (
    BaseTriggerChannel,
    TriggerChannelEvents,
)
from port_ocean.core.trigger_channel.settings import KafkaTriggerChannelSettings


class KafkaTriggerChannel(BaseTriggerChannel):
    def __init__(
        self,
        events: TriggerChannelEvents,
        trigger_channel_config: KafkaTriggerChannelSettings,
        org_id: str,
    ):
        super().__init__(events)
        self.trigger_channel_config = trigger_channel_config
        self.org_id = org_id

    async def _get_kafka_creds(self) -> KafkaConsumerConfig:
        if self.trigger_channel_config.kafka_security_enabled:
            creds = await ocean.port_client.get_kafka_creds()
            return KafkaConsumerConfig(
                username=creds.get("username"),
                password=creds.get("password"),
                brokers=self.trigger_channel_config.brokers,
                security_protocol=self.trigger_channel_config.security_protocol,
                authentication_mechanism=self.trigger_channel_config.authentication_mechanism,
                kafka_security_enabled=self.trigger_channel_config.kafka_security_enabled,
            )

        return KafkaConsumerConfig(
            brokers=self.trigger_channel_config.brokers,
            security_protocol=self.trigger_channel_config.security_protocol,
            authentication_mechanism=self.trigger_channel_config.authentication_mechanism,
            kafka_security_enabled=self.trigger_channel_config.kafka_security_enabled,
        )

    def should_be_processed(self, msg_value: dict[Any, Any], topic: str) -> bool:
        if "runs" in topic:
            return (
                msg_value.get("payload", {})
                .get("action", {})
                .get("invocationMethod", {})
                .get("type", "")
                == "KAFKA"
            )

        if "change.log" in topic:
            return msg_value.get("changelogDestination", {}).get("type", "") == "KAFKA"

        return False

    async def _handle_message(self, message: dict[Any, Any], topic: str) -> None:
        if not self.should_be_processed(message, topic):
            return

        if "change.log" in topic:
            await self.events["on_resync"](message)

        if "runs" in topic:
            await self.events["on_action"](message)

    def wrapped_start(
        self, context: PortOceanContext, func: Callable[[], None]
    ) -> Callable[[], None]:
        ocean_app = context.app

        def wrapper() -> None:
            initialize_port_ocean_context(ocean_app=ocean_app)
            func()

        return wrapper

    async def start(self) -> None:
        consumer = KafkaConsumer(
            msg_process=self._handle_message,
            org_id=self.org_id,
            config=await self._get_kafka_creds(),
        )
        threading.Thread(
            name="ocean_kafka_consumer",
            target=self.wrapped_start(ocean, consumer.start),
        ).start()



from time import time
from typing import Callable, Awaitable
from uuid import uuid4

from fastapi import Request, Response

from loguru import logger

from .context.event import event_context
from .context.ocean import ocean


def get_time(seconds_precision: bool = True) -> float:
    """Return current time as Unix/Epoch timestamp, in seconds.
    :param seconds_precision: if True, return with seconds precision as integer (default).
                              If False, return with milliseconds precision as floating point number of seconds.
    """
    return time() if not seconds_precision else int(time())


def get_uuid() -> str:
    """Return a UUID4 as string"""
    return str(uuid4())


async def request_handler(
    request: Request, call_next: Callable[[Request], Awaitable[Response]]
) -> Response:
    """Middleware used by FastAPI to process each request, featuring:

    - Contextualize request logs with a unique Request ID (UUID4) for each unique request.
    - Catch exceptions during the request handling. Translate custom API exceptions into responses,
      or treat (and log) unexpected exceptions.
    """
    start_time = get_time(seconds_precision=False)
    request_id = get_uuid()

    with logger.contextualize(request_id=request_id):
        logger.bind(url=str(request.url), method=request.method).info("Request started")
        response: Response

        if request.url.path.startswith("integration"):
            async with event_context(""):
                await ocean.integration.port_app_config_handler.get_port_app_config()
                response = await call_next(request)
        else:
            response = await call_next(request)

        end_time = get_time(seconds_precision=False)
        time_elapsed = round(end_time - start_time, 5)
        response.headers["X-Request-ID"] = request_id
        response.headers["X-Process-Time"] = str(time_elapsed)
        logger.bind(
            time_elapsed=time_elapsed, response_status=response.status_code
        ).info("Request ended")
        return response



import asyncio
from typing import (
    Any,
)

from loguru import logger

from port_ocean.clients.port.types import UserAgentType
from port_ocean.context.event import (
    event_context,
    TriggerType,
)
from port_ocean.context.ocean import PortOceanContext
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.integrations.mixins.sync import SyncRawMixin, SyncMixin
from port_ocean.core.models import Entity
from port_ocean.core.trigger_channel.factory import (
    TriggerChannelFactory,
)


class BaseIntegration(SyncRawMixin, SyncMixin):
    def __init__(self, context: PortOceanContext):
        SyncRawMixin.__init__(self)
        SyncMixin.__init__(self)
        self.started = False
        self.context = context
        self.trigger_channel = TriggerChannelFactory(
            context,
            self.context.config.integration.identifier,
            {"on_action": self.trigger_action, "on_resync": self.sync_all},
        )

    async def _sync_new_in_batches(
        self, resource_config: ResourceConfig, user_agent_type: UserAgentType
    ) -> list[Entity]:
        resource, results = await self._get_resource_raw_results(resource_config)

        tasks = []

        batch_size = self.context.config.batch_work_size or len(results)
        for batch in [
            results[i : i + batch_size] for i in range(0, len(results), batch_size)
        ]:
            tasks.append(self._register_resource_raw(resource, batch, user_agent_type))
        entities = await asyncio.gather(*tasks)
        return sum(entities, [])

    async def start(self) -> None:
        logger.info("Starting integration")
        if self.started:
            raise Exception("Integration already started")

        if (
            not self.event_strategy["resync"]
            and self.__class__._on_resync == BaseIntegration._on_resync
        ):
            raise NotImplementedError("on_resync is not implemented")

        await self.initialize_handlers()

        logger.info("Initializing integration at port")
        await self.context.port_client.initiate_integration(
            self.context.config.integration.identifier,
            self.context.config.integration.type,
            self.context.config.trigger_channel.to_request(),
        )

        logger.info("Initializing trigger channel")
        await self.trigger_channel.create_trigger_channel()

        self.started = True

        async with event_context("start", trigger_type="machine"):
            await asyncio.gather(
                *(listener() for listener in self.event_strategy["start"])
            )

    async def trigger_action(self, data: dict[Any, Any]) -> None:
        raise NotImplementedError("trigger_action is not implemented")

    async def sync_all(
        self,
        _: dict[Any, Any] | None = None,
        trigger_type: TriggerType = "machine",
        user_agent_type: UserAgentType = UserAgentType.exporter,
    ) -> None:
        logger.info("Resync was triggered")

        async with event_context("resync", trigger_type=trigger_type):
            app_config = await self.port_app_config_handler.get_port_app_config()

            created_entities = await asyncio.gather(
                *(
                    self._sync_new_in_batches(resource, user_agent_type)
                    for resource in app_config.resources
                )
            )

            await self.transport.delete_non_existing(
                sum(created_entities, []), user_agent_type
            )

            logger.info("Resync was finished")



from typing import Iterable, Any, TypeVar

from port_ocean.core.handlers.manipulation.base import EntityPortDiff
from port_ocean.core.models import Entity


def is_valid_diff_item(item: Any) -> bool:
    return isinstance(item, list) and all([isinstance(i, dict) for i in item] or [True])


def validate_result(result: Any) -> list[dict[Any, Any]]:
    if isinstance(result, list):
        if is_valid_diff_item(result):
            return result
    raise Exception(f"Expected dict, got {type(result)} instead")


def is_same_entity(firs_entity: Entity, second_entity: Entity) -> bool:
    return (
        firs_entity.identifier == second_entity.identifier
        and firs_entity.blueprint == second_entity.blueprint
    )


def get_unique(array: list[Entity]) -> list[Entity]:
    seen: list[Entity] = []
    result = []
    for item in array:
        if all(not is_same_entity(item, seen_item) for seen_item in seen):
            seen.append(item)
            result.append(item)
    return result


T = TypeVar("T", bound=list[Any])


def zip_and_sum(collection: Iterable[tuple[T, ...]]) -> tuple[T, ...]:
    return tuple(sum(items, []) for items in zip(*collection))  # type: ignore


def get_port_diff(
    before: Iterable[Entity],
    after: Iterable[Entity],
) -> EntityPortDiff:
    return EntityPortDiff(
        deleted=get_unique(
            [
                item
                for item in before
                if not any(is_same_entity(item, item_after) for item_after in after)
            ],
        ),
        created=get_unique(
            [
                item
                for item in after
                if not any(is_same_entity(item, item_before) for item_before in before)
            ],
        ),
        modified=get_unique(
            [
                item
                for item in after
                if any(is_same_entity(item, entity_before) for entity_before in before)
            ],
        ),
    )



from typing import (
    Awaitable,
    Callable,
    TypedDict,
    Any,
)

from port_ocean.core.models import Entity


class RawEntityDiff(TypedDict):
    before: list[dict[Any, Any]]
    after: list[dict[Any, Any]]


class EntityDiff(TypedDict):
    before: list[Entity]
    after: list[Entity]


RESYNC_EVENT_LISTENER = Callable[[str], Awaitable[list[dict[Any, Any]]]]
START_EVENT_LISTENER = Callable[[], Awaitable]


class IntegrationEventsCallbacks(TypedDict):
    start: list[START_EVENT_LISTENER]
    resync: dict[str | None, list[RESYNC_EVENT_LISTENER]]



import asyncio
from itertools import chain
from typing import Any, Awaitable

from loguru import logger

from port_ocean.clients.port.types import UserAgentType
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.integrations.mixins.events import EventsMixin
from port_ocean.core.integrations.mixins.handler import HandlerMixin
from port_ocean.core.models import Entity
from port_ocean.core.utils import validate_result, zip_and_sum
from port_ocean.types import RawEntityDiff, EntityDiff


class SyncMixin(HandlerMixin, EventsMixin):
    def __init__(self) -> None:
        HandlerMixin.__init__(self)
        EventsMixin.__init__(self)

    async def register(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType,
    ) -> None:
        await self.transport.upsert(entities, user_agent_type)
        logger.info("Finished registering change")

    async def unregister(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        await self.transport.delete(entities, user_agent_type)
        logger.info("Finished unregistering change")

    async def sync(
        self,
        entities: list[Entity],
        user_agent_type: UserAgentType,
    ) -> None:
        await self.transport.upsert(entities, user_agent_type)
        await self.transport.delete_non_existing(entities, user_agent_type)

        logger.info("Finished syncing change")


class SyncRawMixin(HandlerMixin, EventsMixin):
    def __init__(self) -> None:
        HandlerMixin.__init__(self)
        EventsMixin.__init__(self)

    async def _on_resync(self, kind: str) -> list[dict[Any, Any]]:
        raise NotImplementedError("on_resync must be implemented")

    async def _calculate_raw(
        self, raw_diff: list[tuple[ResourceConfig, RawEntityDiff]]
    ) -> list[EntityDiff]:
        logger.info("Calculating diff in entities between states")
        return await asyncio.gather(
            *[
                self.manipulation.parse_items(mapping, results)
                for mapping, results in raw_diff
            ]
        )

    async def _get_resource_raw_results(
        self, resource_config: ResourceConfig
    ) -> tuple[ResourceConfig, list[dict[Any, Any]]]:
        logger.info(f"Resyncing {resource_config.kind}")
        tasks: list[Awaitable[list[dict[Any, Any]]]] = []
        with logger.contextualize(kind=resource_config.kind):
            if self.__class__._on_resync != SyncRawMixin._on_resync:
                tasks.append(self._on_resync(resource_config.kind))

            fns = [
                *self.event_strategy["resync"][resource_config.kind],
                *self.event_strategy["resync"][None],
            ]

            for wrapper in fns:
                tasks.append(wrapper(resource_config.kind))

            logger.info(f"Found {len(tasks)} resync tasks for {resource_config.kind}")
            results: list[dict[Any, Any]] = list(
                chain.from_iterable(
                    [
                        validate_result(task_result)
                        for task_result in await asyncio.gather(*tasks)
                    ]
                )
            )

            logger.info(f"Triggered {len(tasks)} tasks for {resource_config.kind}")
            return resource_config, results

    async def _register_resource_raw(
        self,
        resource: ResourceConfig,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        objects_diff = await self._calculate_raw(
            [
                (
                    resource,
                    {
                        "before": [],
                        "after": results,
                    },
                )
            ]
        )

        entities_after: list[Entity] = objects_diff[0]["after"]
        await self.transport.upsert(entities_after, user_agent_type)
        logger.info("Finished registering change")
        return entities_after

    async def _unregister_resource_raw(
        self,
        resource: ResourceConfig,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        objects_diff = await self._calculate_raw(
            [
                (
                    resource,
                    {
                        "before": results,
                        "after": [],
                    },
                )
            ]
        )

        entities_after: list[Entity] = objects_diff[0]["before"]
        await self.transport.delete(entities_after, user_agent_type)
        logger.info("Finished unregistering change")
        return entities_after

    async def register_raw(
        self,
        kind: str,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        logger.info(f"Registering state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        return await asyncio.gather(
            *(
                self._register_resource_raw(resource, results, user_agent_type)
                for resource in resource_mappings
            )
        )

    async def unregister_raw(
        self,
        kind: str,
        results: list[dict[Any, Any]],
        user_agent_type: UserAgentType,
    ) -> list[Entity]:
        logger.info(f"Registering state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        return await asyncio.gather(
            *(
                self._unregister_resource_raw(resource, results, user_agent_type)
                for resource in resource_mappings
            )
        )

    async def update_raw_diff(
        self,
        kind: str,
        raw_desired_state: RawEntityDiff,
        user_agent_type: UserAgentType,
    ) -> None:
        logger.info(f"Registering state for {kind}")
        config = await self.port_app_config_handler.get_port_app_config()
        resource_mappings = [
            resource for resource in config.resources if resource.kind == kind
        ]

        with logger.contextualize(kind=kind):
            logger.info(f"Found {len(resource_mappings)} resources for {kind}")

            objects_diff = await self._calculate_raw(
                [(mapping, raw_desired_state) for mapping in resource_mappings]
            )

            entities_before, entities_after = zip_and_sum(
                (
                    (entities_change["before"], entities_change["after"])
                    for entities_change in objects_diff
                )
            )

            await self.transport.update_diff(
                {"before": entities_before, "after": entities_after}, user_agent_type
            )



from importlib.util import find_spec

try:
    find_spec("click")
    find_spec("cookiecutter")
    find_spec("rich")

    cli_included = True
except ImportError:
    cli_included = False

defaults = [
    "clients",
    "config",
    "consumers",
    "context",
    "core",
    "common",
    "logging",
    "port_ocean",
]

if cli_included:
    from port_ocean.cli import cli

    __all__ = [*defaults, "cli"]
else:
    __all__ = defaults



import asyncio
from itertools import groupby

from port_ocean.clients.port.client import PortClient
from port_ocean.core.handlers.manipulation.base import EntityPortDiff
from port_ocean.core.handlers.transport.port.get_required_entities import (
    get_required_entities,
)


async def validate_entity_relations(
    diff: EntityPortDiff, port_client: PortClient
) -> None:
    modified_or_created_entities = diff.modified + diff.created
    entities_with_relations = [
        entity for entity in modified_or_created_entities if entity.relations
    ]
    blueprint_identifier_to_entity = dict(
        groupby(
            entities_with_relations,
            key=lambda x: x.blueprint,
        )
    )
    blueprints = await asyncio.gather(
        *[
            port_client.get_blueprint(blueprint_identifier)
            for blueprint_identifier in blueprint_identifier_to_entity.keys()
        ]
    )
    entity_to_blueprint = [
        (
            entity,
            next(
                blueprint
                for blueprint in blueprints
                if blueprint.identifier == entity.blueprint
            ),
        )
        for entity in entities_with_relations
    ]

    required_entities = get_required_entities(
        entity_to_blueprint, diff.deleted, modified_or_created_entities
    )

    await asyncio.gather(
        *[
            port_client.validate_entity_exist(item.identifier, item.blueprint)
            for item in required_entities
        ]
    )



import requests


def list_git_folders(repo_url: str, path: str) -> list[str]:
    # Parse the repository URL to extract the owner and repository name
    parts = repo_url.split("/")
    owner = parts[-2]
    repo_name = parts[-1].split(".")[0]

    # Construct the API URL to get the contents of the folder
    api_url = f"https://api.github.com/repos/{owner}/{repo_name}/contents/{path}"

    # Send a GET request to the API
    response = requests.get(api_url)

    # Check if the request was successful
    if response.status_code == 200:
        contents = response.json()
        folders = [item["name"] for item in contents if item["type"] == "dir"]
        return folders
    else:
        print(f"Failed to list folders. {response.text}")
        return []



from abc import abstractmethod

from port_ocean.clients.port.types import UserAgentType
from port_ocean.core.base import BaseWithContext
from port_ocean.core.models import Entity
from port_ocean.types import EntityDiff


class BaseTransport(BaseWithContext):
    DEFAULT_USER_AGENT_TYPE = UserAgentType.exporter

    @abstractmethod
    async def update_diff(
        self,
        entities: EntityDiff,
        user_agent: UserAgentType | None = None,
    ) -> None:
        pass

    @abstractmethod
    async def upsert(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        pass

    @abstractmethod
    async def delete(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        pass

    @abstractmethod
    async def delete_non_existing(
        self, entities: list[Entity], user_agent: UserAgentType
    ) -> None:
        pass



import asyncio
from itertools import chain

from loguru import logger

from port_ocean.clients.port.types import UserAgentType
from port_ocean.context.event import event
from port_ocean.core.handlers.manipulation.base import (
    EntityPortDiff,
)
from port_ocean.core.handlers.transport.base import BaseTransport
from port_ocean.core.handlers.transport.port.order_by_entities_dependencies import (
    order_by_entities_dependencies,
)
from port_ocean.core.handlers.transport.port.validate_entity_relations import (
    validate_entity_relations,
)
from port_ocean.core.models import Entity
from port_ocean.core.utils import (
    is_same_entity,
    get_unique,
    get_port_diff,
)
from port_ocean.types import EntityDiff


class HttpPortTransport(BaseTransport):
    async def _validate_delete_dependent_entities(self, entities: list[Entity]) -> None:
        logger.info("Validated deleted entities")
        if not event.port_app_config.delete_dependent_entities:
            deps = await asyncio.gather(
                *[
                    self.context.port_client.search_dependent_entities(entity)
                    for entity in entities
                ]
            )
            new_dependent = get_unique(
                [
                    entity
                    for entity in chain.from_iterable(deps)
                    if not any([is_same_entity(item, entity) for item in entities])
                ]
            )

            if new_dependent:
                raise Exception(
                    f"Must enable delete_dependent_entities flag or delete also dependent entities:"
                    f" {[(dep.blueprint, dep.identifier) for dep in new_dependent]}"
                )

    async def _validate_entity_diff(self, diff: EntityPortDiff) -> None:
        config = event.port_app_config
        await self._validate_delete_dependent_entities(diff.deleted)
        modified_or_created_entities = diff.modified + diff.created
        logger.info("Validating modified or created entities")

        await asyncio.gather(
            *[
                self.context.port_client.validate_entity_payload(
                    entity,
                    {
                        "merge": config.enable_merge_entity,
                        "create_missing_related_entities": config.create_missing_related_entities,
                    },
                )
                for entity in modified_or_created_entities
            ]
        )
        await validate_entity_relations(diff, self.context.port_client)

    async def update_diff(
        self,
        entities: EntityDiff,
        user_agent_type: UserAgentType | None = None,
    ) -> None:
        diff = get_port_diff(entities["before"], entities["after"])

        logger.info(
            f"Registering entity diff (created: {len(diff.created)}, deleted: {len(diff.deleted)}, modified: {len(diff.modified)})"
        )
        await self._validate_entity_diff(diff)

        user_agent_type = user_agent_type or self.DEFAULT_USER_AGENT_TYPE
        await self.delete(diff.deleted, user_agent_type)
        await self.upsert(diff.created, user_agent_type)
        await self.upsert(diff.modified, user_agent_type)

    async def upsert(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        ordered_created_entities = reversed(order_by_entities_dependencies(entities))
        for entity in ordered_created_entities:
            await self.context.port_client.upsert_entity(
                entity,
                event.port_app_config.get_port_request_options(),
                user_agent_type,
            )

    async def delete(
        self, entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        ordered_deleted_entities = order_by_entities_dependencies(entities)

        await asyncio.gather(
            *[
                self.context.port_client.delete_entity(entity, user_agent_type)
                for entity in ordered_deleted_entities
            ]
        )

    async def delete_non_existing(
        self, excluded_entities: list[Entity], user_agent_type: UserAgentType
    ) -> None:
        entities_at_port = await self.context.port_client.search_entities(
            user_agent_type
        )
        diff = get_port_diff(entities_at_port, excluded_entities)
        await self._validate_entity_diff(diff)

        ordered_deleted_entities = order_by_entities_dependencies(diff.deleted)

        await asyncio.gather(
            *[
                self.context.port_client.delete_entity(entity, user_agent_type)
                for entity in ordered_deleted_entities
            ]
        )



from typing import Callable, Any, Awaitable

from port_ocean.context.ocean import PortOceanContext
from port_ocean.core.base import BaseWithContext
from port_ocean.core.trigger_channel.base import (
    BaseTriggerChannel,
    TriggerChannelEvents,
)
from port_ocean.core.trigger_channel.http import (
    HttpTriggerChannel,
)
from port_ocean.core.trigger_channel.kafka import (
    KafkaTriggerChannel,
)
from port_ocean.core.trigger_channel.settings import (
    HttpTriggerChannelSettings,
    KafkaTriggerChannelSettings,
)


class TriggerChannelFactory(BaseWithContext):
    def __init__(
        self,
        context: PortOceanContext,
        installation_id: str,
        events: TriggerChannelEvents,
    ):
        super().__init__(context)
        self.installation_id = installation_id
        self._trigger_channel: BaseTriggerChannel | None = None
        self.events = events

    def on_event(
        self, callback: Callable[[dict[Any, Any]], Awaitable[None]]
    ) -> Callable[[dict[Any, Any]], Awaitable[None]]:
        async def wrapper(event: dict[Any, Any]) -> None:
            integration_identifier = (
                event.get("diff", {}).get("after", {}).get("identifier")
            )

            if integration_identifier == self.installation_id:
                await callback(event)

        return wrapper

    async def create_trigger_channel(self) -> None:
        wrapped_events: TriggerChannelEvents = {
            "on_resync": self.on_event(self.events["on_resync"]),
            "on_action": self.on_event(self.events["on_action"]),
        }
        config = self.context.config.trigger_channel
        _type = config.type.lower()
        assert_message = "Invalid trigger channel config, expected KafkaTriggerChannelSettings and got {0}"

        match _type:
            case "kafka":
                assert isinstance(
                    config, KafkaTriggerChannelSettings
                ), assert_message.format(type(config))
                org_id = await self.context.port_client.get_org_id()
                self._trigger_channel = KafkaTriggerChannel(
                    wrapped_events,
                    config,
                    org_id,
                )

            case "webhook":
                assert isinstance(
                    config, HttpTriggerChannelSettings
                ), assert_message.format(type(config))
                self._trigger_channel = HttpTriggerChannel(wrapped_events, config)

            case _:
                raise Exception(f"Trigger channel {_type} not supported")

        await self._trigger_channel.start()



from typing import Literal, Any

from pydantic import AnyHttpUrl, Field, BaseSettings


class TriggerChannelSettings(BaseSettings):
    type: str

    def to_request(self) -> dict[str, Any]:
        return {"type": self.type}


class HttpTriggerChannelSettings(TriggerChannelSettings):
    type: Literal["WEBHOOK"]
    app_host: AnyHttpUrl = Field(alias="appHost")

    def to_request(self) -> dict[str, Any]:
        return {
            **super().to_request(),
            "url": self.app_host + "/resync",
        }


class KafkaTriggerChannelSettings(TriggerChannelSettings):
    type: Literal["KAFKA"]
    brokers: str = ""
    security_protocol: str = Field(alias="securityProtocol", default="SASL_SSL")
    authentication_mechanism: str = Field(
        alias="authenticationMechanism", default="SCRAM-SHA-512"
    )
    kafka_security_enabled: bool = Field(alias="kafkaSecurityEnabled", default=True)



from abc import abstractmethod
from dataclasses import dataclass, field

from port_ocean.core.base import BaseWithContext
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.models import Entity
from port_ocean.types import RawEntityDiff, EntityDiff


@dataclass
class EntityPortDiff:
    deleted: list[Entity] = field(default_factory=list)
    modified: list[Entity] = field(default_factory=list)
    created: list[Entity] = field(default_factory=list)


class BaseManipulation(BaseWithContext):
    @abstractmethod
    async def parse_items(
        self, mapping: ResourceConfig, raw_data: RawEntityDiff
    ) -> EntityDiff:
        pass



import os
import re
from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseSettings


def read_yaml_config_settings_source(
    settings: "BaseOceanSettings", base_path: str
) -> str:
    """Loads settings from a YAML file at `Config.yaml_file`

    "<file:xxxx>" patterns are replaced with the contents of file xxxx. The root path
    were to find the files is configured with `secrets_dir`.
    """
    yaml_file = getattr(settings.__config__, "yaml_file", "")

    assert yaml_file, "Settings.yaml_file not properly configured"

    path = Path(base_path) / yaml_file

    if not path.exists():
        raise FileNotFoundError(f"Could not open yaml settings file at: {path}")

    return path.read_text("utf-8")


def validate_config_provider_pattern(value: str) -> tuple[str, str]:
    pattern = r"^[a-zA-Z0-9]+ .*$"
    match = re.match(pattern, value)
    if not match:
        raise ValueError(f"Invalid pattern: {value}. Pattern should match: {pattern}")

    index = value.find(" ")
    provider_type, provider_value = value[:index], value[index + 1 :]

    return provider_type, provider_value


def is_config_provider_syntax(value: str) -> bool:
    pattern = r"^{{ from .* }}$"
    match = re.match(pattern, value)
    if match:
        return True
    else:
        return False


def load_from_config_provider(provider_type: str, value: str) -> Any:
    if provider_type == "env":
        result = os.environ.get(value)
        if result is None:
            raise ValueError(f"Environment variable not found: {value}")
        return result
    else:
        raise ValueError(f"Invalid provider type: {provider_type}")


def _from_provider(settings: "BaseOceanSettings", base_path: str) -> dict[str, Any]:
    value = read_yaml_config_settings_source(settings, base_path)
    pattern = r"\{\{ from (.*) \}\}"
    matches = re.finditer(pattern, value)
    for match in matches:
        provider_type, provider_value = validate_config_provider_pattern(match.group(1))
        data = load_from_config_provider(provider_type, provider_value)
        value = re.sub(re.escape(match.group()), data, value, count=1)

    return yaml.safe_load(value)


class BaseOceanSettings(BaseSettings):
    base_path: str

    class Config:
        secrets_dir = "./secrets.yml"
        yaml_file = "./config.yaml"

        @classmethod
        def customise_sources(cls, init_settings, *_, **__):  # type: ignore
            return (
                init_settings,
                lambda s: _from_provider(s, init_settings.init_kwargs["base_path"]),
            )



# -*- coding: utf-8 -*-
# ruff: noqa: E501

import os

import click
from cookiecutter.main import cookiecutter  # type: ignore
from rich import print
from rich.console import Console

from port_ocean.cli.download_git_folder import download_folder
from port_ocean.cli.list_integrations import list_git_folders


def print_logo() -> None:
    ascii_art = """
=====================================================================================
          ::::::::       ::::::::       ::::::::::           :::        ::::    ::: 
        :+:    :+:     :+:    :+:      :+:                :+: :+:      :+:+:   :+:  
       +:+    +:+     +:+             +:+               +:+   +:+     :+:+:+  +:+   
      +#+    +:+     +#+             +#++:++#         +#++:++#++:    +#+ +:+ +#+    
     +#+    +#+     +#+             +#+              +#+     +#+    +#+  +#+#+#     
    #+#    #+#     #+#    #+#      #+#              #+#     #+#    #+#   #+#+#      
    ########       ########       ##########       ###     ###    ###    ####      
=====================================================================================
By: Port.io
"""

    # Display ASCII art
    Console().print(ascii_art)


@click.group()
def cli_start() -> None:
    # Ocean root command
    pass


@cli_start.command()
@click.argument("path", default="")
def sail(path: str) -> None:
    from port_ocean.port_ocean import run

    print_logo()

    print("Setting sail... ⛵️⚓️⛵️⚓️ All hands on deck! ⚓️")
    run(path)


@cli_start.command()
@click.argument("path", default=".")
def new(path: str) -> None:
    print_logo()

    console = Console()
    console.print(
        "🚢 Unloading cargo... Setting up your integration at the port.", style="bold"
    )

    result = cookiecutter(f"{os.path.dirname(__file__)}/cookiecutter", output_dir=path)
    name = result.split("/")[-1]

    console.print(
        "\n🌊 Ahoy, Captain! Your project has set sail into the vast ocean of possibilities!",
        style="bold",
    )
    console.print("Here are your next steps: \n", style="bold")
    console.print(
        "⚓️ Install necessary packages: Run [bold][blue]make install[/blue][/bold] to install all required packages for your project.\n"
        f"▶️ [bold][blue]cd {path}/{name} && make install && . venv/bin/activate[/blue][/bold]\n"
    )
    console.print(
        "⚓️ Set sail with [blue]Ocean[/blue]: Run [bold][blue]ocean sail[/blue] <path_to_integration>[/bold] to run the project using Ocean.\n"
        f"▶️ [bold][blue]ocean sail {path}/{name}[/blue][/bold] \n"
    )
    console.print(
        "⚓️ Smooth sailing with [blue]Make[/blue]: Alternatively, you can run [bold][blue]make run[/blue][/bold] to launch your project using Make. \n"
        f"▶️ [bold][blue]make run {path}/{name}[/blue][/bold]"
    )


@cli_start.command(name="list")
def list_integrations() -> None:
    console = Console()
    console.print("🌊 Here are the integrations available to you:", style="bold")
    options = list_git_folders("https://github.com/port-labs/pulumi", "examples")

    for option in options:
        console.print(f"⚓️ [bold][blue]{option}[/blue][/bold]")


@cli_start.command()
@click.argument("name")
def pull(name: str) -> None:
    download_folder(
        "https://github.com/port-labs/pulumi", f"examples/{name}", f"./{name}"
    )



from fastapi import APIRouter

from port_ocean.context.ocean import ocean
from port_ocean.core.trigger_channel.base import (
    BaseTriggerChannel,
    TriggerChannelEvents,
)
from port_ocean.core.trigger_channel.settings import HttpTriggerChannelSettings


class HttpTriggerChannel(BaseTriggerChannel):
    def __init__(
        self,
        events: TriggerChannelEvents,
        trigger_channel_config: HttpTriggerChannelSettings,
    ):
        super().__init__(events)
        self.trigger_channel_config = trigger_channel_config

    async def start(self) -> None:
        target_channel_router = APIRouter()

        @target_channel_router.post("/resync")
        async def resync() -> None:
            await self.events["on_resync"]({})

        ocean.app.fast_api_app.include_router(target_channel_router)



import warnings
from typing import Any

from pydantic import BaseModel, Field
from pydantic import BaseSettings

from port_ocean.config.base import BaseOceanSettings
from port_ocean.core.trigger_channel.settings import (
    HttpTriggerChannelSettings,
    KafkaTriggerChannelSettings,
)

warnings.filterwarnings("ignore", category=FutureWarning)


class PortSettings(BaseSettings):
    client_id: str = Field(alias="clientId")
    client_secret: str = Field(alias="clientSecret")
    base_url: str = Field(alias="baseUrl")


class IntegrationSettings(BaseSettings):
    identifier: str
    type: str
    config: dict[str, Any]


class IntegrationConfiguration(BaseOceanSettings):
    port: PortSettings
    trigger_channel: KafkaTriggerChannelSettings | HttpTriggerChannelSettings = Field(
        alias="triggerChannel"
    )
    batch_work_size: int | None = Field(alias="batchWorkSize", default=None)
    integration: IntegrationSettings


class LoggerConfiguration(BaseModel):
    level: str = "DEBUG"
    serialize: bool = False



from port_ocean.core.models import Entity, Blueprint
from port_ocean.core.utils import is_same_entity


def get_required_entities(
    entity_to_blueprint: list[tuple[Entity, Blueprint]],
    forbidden_entities: list[Entity],
    excluded_entities: list[Entity],
) -> list[Entity]:
    required_entities = []
    for entity, blueprint in entity_to_blueprint:
        for relation_name, relation in entity.relations.items():
            relation_blueprint = blueprint.relations[relation_name].target
            target_entity = Entity(identifier=relation, blueprint=relation_blueprint)

            if any(is_same_entity(item, target_entity) for item in forbidden_entities):
                raise Exception(
                    f"Cant delete entity {target_entity} of blueprint {target_entity.blueprint} "
                    f"because it was specified as relation target of entity {entity} "
                    f"of blueprint {entity.blueprint}"
                )

            if not any(
                is_same_entity(item, target_entity) for item in excluded_entities
            ):
                required_entities.append(target_entity)

    return required_entities



from abc import abstractmethod
from typing import Type, Any

from port_ocean.context.event import event
from port_ocean.core.base import BaseWithContext
from port_ocean.core.handlers.port_app_config.models import PortAppConfig


class BasePortAppConfig(BaseWithContext):
    CONFIG_CLASS: Type[PortAppConfig] = PortAppConfig

    @abstractmethod
    async def _get_port_app_config(self) -> dict[str, Any]:
        pass

    async def get_port_app_config(self) -> PortAppConfig:
        raw_config = await self._get_port_app_config()
        config = self.CONFIG_CLASS.parse_obj(raw_config)
        event.port_app_config = config
        return config



from datetime import datetime
from typing import Any

import httpx as httpx
from loguru import logger
from pydantic import BaseModel, Field, PrivateAttr

from port_ocean.clients.port.types import (
    KafkaCreds,
    RequestOptions,
    UserAgentType,
)
from port_ocean.core.models import Entity, Blueprint


class TokenResponse(BaseModel):
    access_token: str = Field(alias="accessToken")
    expires_in: int = Field(alias="expiresIn")
    token_type: str = Field(alias="tokenType")
    _retrieved_time: datetime = PrivateAttr(datetime.now())

    @property
    def expired(self) -> bool:
        return (
            self._retrieved_time.timestamp() + self.expires_in
            < datetime.now().timestamp()
        )

    @property
    def full_token(self) -> str:
        return f"{self.token_type} {self.access_token}"


class PortClient:
    def __init__(
        self, base_url: str, client_id: str, client_secret: str, user_agent_id: str
    ):
        self.api_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.user_agent_id = user_agent_id
        self._last_token_object: TokenResponse | None = None

    async def _get_token(self, client_id: str, client_secret: str) -> TokenResponse:
        async with httpx.AsyncClient() as client:
            logger.info(f"Fetching access token for clientId: {client_id}")

            credentials = {"clientId": client_id, "clientSecret": client_secret}
            token_response = await client.post(
                f"{self.api_url}/auth/access_token", json=credentials
            )
            token_response.raise_for_status()
            return TokenResponse(**token_response.json())

    def _user_agent(self, user_agent_type: UserAgentType | None = None) -> str:
        user_agent = f"port-ocean/{self.user_agent_id}"
        if user_agent_type:
            user_agent += f"/{user_agent_type.value or UserAgentType.exporter.value}"

        return user_agent

    async def _headers(
        self, user_agent_type: UserAgentType | None = None
    ) -> dict[Any, Any]:
        return {
            "Authorization": await self.token,
            "User-Agent": self._user_agent(user_agent_type),
        }

    @property
    async def token(self) -> str:
        logger.info("Fetching access token")
        if not self._last_token_object or self._last_token_object.expired:
            self._last_token_object = await self._get_token(
                self.client_id, self.client_secret
            )
        else:
            logger.info("Access token found in cache")

        return self._last_token_object.full_token

    async def upsert_entity(
        self,
        entity: Entity,
        request_options: RequestOptions,
        user_agent_type: UserAgentType | None = None,
    ) -> None:
        validation_only = request_options.get("validation_only", False)
        logger.info(
            f"{'Validating' if validation_only else 'Upserting'} entity: {entity.identifier} of blueprint: {entity.blueprint}"
        )
        headers = await self._headers(user_agent_type)

        async with httpx.AsyncClient() as client:
            response = await client.post(
                f"{self.api_url}/blueprints/{entity.blueprint}/entities",
                json=entity.dict(exclude_unset=True),
                headers=headers,
                params={
                    "upsert": "true",
                    "merge": str(request_options.get("merge", False)).lower(),
                    "create_missing_related_entities": str(
                        request_options.get("create_missing_related_entities", False)
                    ).lower(),
                    "validation_only": str(validation_only).lower(),
                },
            )

        if not response.status_code < 400:
            logger.error(
                f"Error upserting "
                f"entity: {entity.identifier} of "
                f"blueprint: {entity.blueprint}, "
                f"error: {response.text}"
            )
        response.raise_for_status()

    async def delete_entity(
        self, entity: Entity, user_agent_type: UserAgentType | None = None
    ) -> None:
        logger.info(
            f"Delete entity: {entity.identifier} of blueprint: {entity.blueprint}"
        )
        async with httpx.AsyncClient() as client:
            response = await client.delete(
                f"{self.api_url}/blueprints/{entity.blueprint}/entities/{entity.identifier}",
                headers=await self._headers(user_agent_type),
                params={"delete_dependents": "true"},
            )

        if not response.status_code < 400:
            logger.error(
                f"Error deleting "
                f"entity: {entity.identifier} of "
                f"blueprint: {entity.blueprint}, "
                f"error: {response.text}"
            )
            response.raise_for_status()

    async def get_kafka_creds(self) -> KafkaCreds:
        logger.info("Fetching organization kafka credentials")
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self.api_url}/kafka-credentials", headers=await self._headers()
            )
        if not response.status_code < 400:
            logger.error(f"Error getting kafka credentials, error: {response.text}")
            response.raise_for_status()

        credentials = response.json()["credentials"]

        if credentials is None:
            raise Exception("No kafka credentials found")

        return credentials

    async def get_org_id(self) -> str:
        logger.info("Fetching organization id")

        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self.api_url}/organization", headers=await self._headers()
            )
        if not response.status_code < 400:
            logger.error(f"Error getting organization id, error: {response.text}")
            response.raise_for_status()

        return response.json()["organization"]["id"]

    async def validate_entity_exist(self, identifier: str, blueprint: str) -> None:
        logger.info(f"Validating entity {identifier} of blueprint {blueprint} exists")

        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self.api_url}/blueprints/{blueprint}/entities/{identifier}",
                headers=await self._headers(),
            )
        if response.status_code >= 400:
            logger.error(
                f"Error validating "
                f"entity: {identifier} of "
                f"blueprint: {blueprint}, "
                f"error: {response.text}"
            )
        response.raise_for_status()

    async def search_entities(self, user_agent_type: UserAgentType) -> list[Entity]:
        query = {
            "combinator": "and",
            "rules": [
                {
                    "property": "$datasource",
                    "operator": "=",
                    "value": self._user_agent(user_agent_type),
                },
            ],
        }

        async with httpx.AsyncClient() as client:
            search_req = await client.post(
                f"{self.api_url}/entities/search",
                json=query,
                headers=await self._headers(),
                params={
                    "exclude_calculated_properties": "true",
                    "include": ["blueprint", "identifier"],
                },
            )
            search_req.raise_for_status()
            return [
                Entity.parse_obj(result) for result in search_req.json()["entities"]
            ]

    async def search_dependent_entities(self, entity: Entity) -> list[Entity]:
        body = {
            "combinator": "and",
            "rules": [
                {
                    "operator": "relatedTo",
                    "blueprint": entity.blueprint,
                    "value": entity.identifier,
                    "direction": "downstream",
                }
            ],
        }

        logger.info(f"Search dependent entity with body {body}")
        async with httpx.AsyncClient() as client:
            response = await client.post(
                f"{self.api_url}/entities/search",
                headers=await self._headers(),
                json=body,
            )
        response.raise_for_status()

        return [Entity.parse_obj(result) for result in response.json()["entities"]]

    async def validate_entity_payload(
        self, entity: Entity, options: RequestOptions
    ) -> None:
        logger.info(f"Validating entity {entity.identifier}")
        await self.upsert_entity(
            entity,
            {
                "merge": options.get("merge", False),
                "create_missing_related_entities": options.get(
                    "create_missing_related_entities", False
                ),
                "validation_only": True,
            },
        )

    async def get_integration(self, identifier: str) -> dict[str, Any]:
        logger.info(f"Fetching integration with id: {identifier}")
        async with httpx.AsyncClient() as client:
            integration = await client.get(
                f"{self.api_url}/integration/{identifier}",
                headers=await self._headers(),
            )
        integration.raise_for_status()
        return integration.json()["integration"]

    async def initiate_integration(
        self, _id: str, _type: str, changelog_destination: dict[str, Any]
    ) -> None:
        logger.info(f"Initiating integration with id: {_id}")
        headers = await self._headers()
        json = {
            "installationId": _id,
            "installationAppType": _type,
            "changelogDestination": changelog_destination,
        }
        async with httpx.AsyncClient() as client:
            installation = await client.patch(
                f"{self.api_url}/integration/{_id}",
                headers=headers,
                json=json,
            )
            if installation.status_code == 404:
                installation = await client.post(
                    f"{self.api_url}/integration", headers=headers, json=json
                )

        if installation.status_code >= 400:
            logger.error(
                f"Error initiating integration with id: {_id}, error: {installation.text}"
            )

        installation.raise_for_status()
        logger.info(f"Integration with id: {_id} successfully registered")

    async def get_blueprint(self, identifier: str) -> Blueprint:
        logger.info(f"Fetching blueprint with id: {identifier}")
        async with httpx.AsyncClient() as client:
            response = await client.get(
                f"{self.api_url}/blueprints/{identifier}", headers=await self._headers()
            )
        response.raise_for_status()
        return Blueprint.parse_obj(response.json()["blueprint"])



import sys

from loguru import logger

from port_ocean.config.integration import LoggerConfiguration

settings = LoggerConfiguration()

LoggerFormat = (
    "<green>{time:YY-MM-DD HH:mm:ss.SSS}</green> | "
    "<level>{level: <8}</level> | "
    "<level>{message}</level> | {extra}"
)

logger.remove()
logger.add(
    sys.stderr,
    level=settings.level.upper(),
    format=LoggerFormat,
    serialize=settings.serialize,
    enqueue=True,  # process logs in background
    diagnose=False,  # hide variable values in log backtrace
)



import sys
from importlib.util import spec_from_file_location, module_from_spec
from inspect import getmembers, isclass
from types import ModuleType
from typing import Type, Callable

import uvicorn
from fastapi import FastAPI, APIRouter
from loguru import logger
from starlette.types import Scope, Receive, Send

from port_ocean.clients.port.client import PortClient
from port_ocean.config.integration import IntegrationConfiguration
from port_ocean.context.ocean import (
    PortOceanContext,
    ocean,
    initialize_port_ocean_context,
)
from port_ocean.core.integrations.base import BaseIntegration
from pydantic import BaseSettings

from port_ocean.middlewares import request_handler


def _get_base_integration_class_from_module(
    module: ModuleType,
) -> Type[BaseIntegration]:
    for name, obj in getmembers(module):
        if (
            isclass(obj)
            and type(obj) == type
            and issubclass(obj, BaseIntegration)
            and obj != BaseIntegration
        ):
            return obj

    raise Exception(f"Failed to load integration from module: {module.__name__}")


def _load_module(file_path: str) -> ModuleType:
    spec = spec_from_file_location("module_name", file_path)
    if spec is None or spec.loader is None:
        raise Exception(f"Failed to load integration from path: {file_path}")

    module = module_from_spec(spec)

    try:
        spec.loader.exec_module(module)
    except Exception as e:
        logger.error(
            f"Failed to load integration with error: {e}, "
            f"please validate the integration type exists"
        )
        raise e

    return module


def _include_target_channel_router(app: FastAPI, _ocean: PortOceanContext) -> None:
    target_channel_router = APIRouter()

    @target_channel_router.post("/resync")
    async def resync() -> None:
        if _ocean.integration is None:
            raise Exception("Integration not set")

        await _ocean.integration.sync_all()

    app.include_router(target_channel_router)


class Ocean:
    def __init__(
        self,
        app: FastAPI | None = None,
        integration_class: Callable[[PortOceanContext], BaseIntegration] | None = None,
        integration_router: APIRouter | None = None,
        config_class: Type[BaseSettings] | None = None,
    ):
        initialize_port_ocean_context(self)
        self.fast_api_app = app or FastAPI()
        self.fast_api_app.middleware("http")(request_handler)

        self.config = IntegrationConfiguration(base_path="./")
        if config_class:
            self.config.integration.config = config_class(
                **self.config.integration.config
            ).dict()
        self.integration_router = integration_router or APIRouter()

        self.port_client = PortClient(
            base_url=self.config.port.base_url,
            client_id=self.config.port.client_id,
            client_secret=self.config.port.client_secret,
            user_agent_id=self.config.integration.identifier,
        )
        self.integration = (
            integration_class(ocean) if integration_class else BaseIntegration(ocean)
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        self.fast_api_app.include_router(self.integration_router, prefix="/integration")
        if self.config.trigger_channel.type == "http":
            _include_target_channel_router(self.fast_api_app, ocean)

        @self.fast_api_app.on_event("startup")
        async def startup() -> None:
            try:
                await self.integration.start()
            except Exception as e:
                logger.error(f"Failed to start integration with error: {e}")
                sys.exit("Server stopped")

        await self.fast_api_app(scope, receive, send)


def run(path: str) -> None:
    sys.path.append(".")
    try:
        integration_path = f"{path}/integration.py" if path else "integration.py"
        module = _load_module(integration_path)
        integration_class = _get_base_integration_class_from_module(module)
    except Exception:
        integration_class = None

    default_app = Ocean(integration_class=integration_class)

    main_path = f"{path}/main.py" if path else "main.py"
    app_module = _load_module(main_path)
    app = {name: item for name, item in getmembers(app_module)}.get("app", default_app)

    uvicorn.run(app, host="0.0.0.0", port=8000)



from functools import lru_cache
from typing import Any

import pyjq as jq  # type: ignore

from port_ocean.core.handlers.manipulation.base import BaseManipulation
from port_ocean.core.handlers.port_app_config.models import ResourceConfig
from port_ocean.core.models import Entity
from port_ocean.types import RawEntityDiff, EntityDiff


class JQManipulation(BaseManipulation):
    @lru_cache
    def _compile(self, pattern: str) -> Any:
        return jq.compile(pattern)

    def _search(self, data: dict[str, Any], pattern: str) -> Any:
        try:
            return self._compile(pattern).first(data) or None
        except Exception:
            return None

    def _search_as_bool(self, data: dict[str, Any], pattern: str) -> bool:
        value = self._compile(pattern).first(data)

        if isinstance(value, bool):
            return value

        raise Exception(f"Expected boolean value, got {type(value)} instead")

    def _search_as_object(
        self, data: dict[str, Any], obj: dict[str, Any]
    ) -> dict[str, Any | None]:
        result: dict[str, Any | None] = {}
        for key, value in obj.items():
            try:
                if isinstance(value, dict):
                    result[key] = self._search_as_object(data, value)
                else:
                    result[key] = self._search(data, value)
            except Exception:
                result[key] = None
        return result

    def _parse_items(
        self, mapping: ResourceConfig, raw_data: list[dict[str, Any]]
    ) -> list[Entity]:
        entities = []
        for data in raw_data:
            should_run = self._search_as_bool(data, mapping.selector.query)

            if should_run and mapping.port.entity:
                entities.append(
                    self._search_as_object(data, mapping.port.entity.mappings.dict())
                )

        return [
            Entity.parse_obj(entity_data)
            for entity_data in filter(
                lambda entity: entity.get("identifier") and entity.get("blueprint"),
                entities,
            )
        ]

    async def parse_items(
        self, mapping: ResourceConfig, raw_results: RawEntityDiff
    ) -> EntityDiff:
        entities_before: list[Entity] = self._parse_items(
            mapping, raw_results["before"]
        )
        entities_after: list[Entity] = self._parse_items(mapping, raw_results["after"])

        return {
            "before": entities_before,
            "after": entities_after,
        }



from typing import Any

from pydantic import BaseModel
from pydantic.fields import Field


class Entity(BaseModel):
    identifier: str
    blueprint: str
    title: str | None
    team: str | list[str] = []
    properties: dict[str, Any] = {}
    relations: dict[str, str] = {}


class BlueprintRelation(BaseModel):
    many: bool
    required: bool
    target: str
    title: str | None


class Blueprint(BaseModel):
    identifier: str
    title: str | None
    team: str | None
    properties_schema: dict[str, Any] = Field(alias="schema")
    relations: dict[str, BlueprintRelation]



from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from typing import AsyncIterator, Literal, Any, TYPE_CHECKING, Optional

from werkzeug.local import LocalStack, LocalProxy

from port_ocean.errors import EventContextNotFoundError

if TYPE_CHECKING:
    from port_ocean.core.handlers.port_app_config.models import PortAppConfig

TriggerType = Literal["manual", "machine"]


@dataclass
class EventContext:
    event_type: str
    trigger_type: TriggerType = "machine"
    attributes: dict[str, Any] = field(default_factory=dict)
    _port_app_config: Optional["PortAppConfig"] = field(default=None)

    @property
    def port_app_config(self) -> "PortAppConfig":
        if self._port_app_config is None:
            raise ValueError("Port app config is not set")
        return self._port_app_config

    @port_app_config.setter
    def port_app_config(self, value: "PortAppConfig") -> None:
        self._port_app_config = value


_event_context_stack: LocalStack[EventContext] = LocalStack()


def _get_event_context() -> EventContext:
    """
    Get the event context from the current thread.
    """
    top_event_context = _event_context_stack.top
    if top_event_context is not None:
        return top_event_context

    raise EventContextNotFoundError(
        "You must be inside an event context in order to use it"
    )


event: EventContext = LocalProxy(lambda: _get_event_context())  # type: ignore


@asynccontextmanager
async def event_context(
    kind: str,
    trigger_type: TriggerType = "manual",
    attributes: dict[str, Any] | None = None,
) -> AsyncIterator[EventContext]:
    if attributes is None:
        attributes = {}

    _event_context_stack.push(
        EventContext(
            kind,
            trigger_type=trigger_type,
            attributes=attributes,
        )
    )

    yield event

    _event_context_stack.pop()



import requests
import os
import shutil


def download_folder(repo_url: str, folder_path: str, destination_path: str) -> None:
    # Parse the repository URL to extract the owner and repository name
    parts = repo_url.split("/")
    owner = parts[-2]
    repo_name = parts[-1].split(".")[0]

    # Construct the API URL to get the contents of the folder
    api_url = f"https://api.github.com/repos/{owner}/{repo_name}/contents/{folder_path}"

    # Send a GET request to the API
    response = requests.get(api_url)

    # Check if the request was successful
    if response.status_code == 200:
        # Create the destination folder if it doesn't exist
        if not os.path.exists(destination_path):
            os.makedirs(destination_path)

        # Iterate over the files and download them
        files = response.json()
        for file in files:
            if file["type"] == "file":
                file_url = file["download_url"]
                file_name = os.path.join(destination_path, file["name"])

                # Download the file
                with requests.get(file_url, stream=True) as r:
                    with open(file_name, "wb") as f:
                        shutil.copyfileobj(r.raw, f)

        print("Folder downloaded successfully!")
    else:
        print("Failed to download the folder.")



from abc import abstractmethod
from typing import TypedDict, Callable, Any, Awaitable


class TriggerChannelEvents(TypedDict):
    on_resync: Callable[[dict[Any, Any]], Awaitable[None]]
    on_action: Callable[[dict[Any, Any]], Awaitable[None]]


class BaseTriggerChannel:
    def __init__(
        self,
        events: TriggerChannelEvents,
    ):
        self.events = events

    @abstractmethod
    async def start(self) -> None:
        pass



from typing import Callable

from loguru import logger

from port_ocean.context.ocean import PortOceanContext, ocean
from port_ocean.core.handlers import BaseManipulation, BasePortAppConfig, BaseTransport
from port_ocean.core.handlers.manipulation.jq_manipulation import JQManipulation
from port_ocean.core.handlers.port_app_config.api import APIPortAppConfig
from port_ocean.core.handlers.transport.port.transport import HttpPortTransport


class HandlerMixin:
    ManipulationHandlerClass: Callable[
        [PortOceanContext], BaseManipulation
    ] = JQManipulation

    AppConfigHandlerClass: Callable[
        [PortOceanContext], BasePortAppConfig
    ] = APIPortAppConfig

    TransportHandlerClass: Callable[
        [PortOceanContext], BaseTransport
    ] = HttpPortTransport

    def __init__(self) -> None:
        self._manipulation: BaseManipulation | None = None
        self._port_app_config_handler: BasePortAppConfig | None = None
        self._transport: BaseTransport | None = None

    @property
    def manipulation(self) -> BaseManipulation:
        if not self._manipulation:
            raise Exception("Integration not started")
        return self._manipulation

    @property
    def port_app_config_handler(self) -> BasePortAppConfig:
        if self._port_app_config_handler is None:
            raise Exception("Integration not started")
        return self._port_app_config_handler

    @property
    def transport(self) -> BaseTransport:
        if not self._transport:
            raise Exception("Integration not started")
        return self._transport

    async def _init_manipulation_instance(self) -> BaseManipulation:
        self._manipulation = self.ManipulationHandlerClass(ocean)
        return self._manipulation

    async def _init_port_app_config_handler_instance(
        self,
    ) -> BasePortAppConfig:
        self._port_app_config_handler = self.AppConfigHandlerClass(ocean)
        return self._port_app_config_handler

    async def _init_transport_instance(self) -> BaseTransport:
        self._transport = self.TransportHandlerClass(ocean)
        return self._transport

    async def initialize_handlers(self) -> None:
        logger.info("Initializing integration components")
        await self._init_manipulation_instance()
        await self._init_port_app_config_handler_instance()
        await self._init_transport_instance()



from graphlib import TopologicalSorter
from typing import Set

from port_ocean.core.models import Entity

Node = tuple[str, str]


def order_by_entities_dependencies(entities: list[Entity]) -> list[Entity]:
    nodes: dict[Node, Set[Node]] = {}
    entities_map = {}

    for entity in entities:
        nodes[(entity.identifier, entity.blueprint)] = set()
        entities_map[(entity.identifier, entity.blueprint)] = entity

    for entity in entities:
        relation_target_ids = [
            identifier
            for relation in entity.relations.values()
            for identifier in relation
        ]
        related_entities = [
            related for related in entities if related.identifier in relation_target_ids
        ]

        for related_entity in related_entities:
            nodes[(entity.identifier, entity.blueprint)].add(
                (related_entity.identifier, related_entity.blueprint)
            )

    sort_op = TopologicalSorter(nodes)
    return [entities_map[item] for item in sort_op.static_order()]



from typing import Any

from port_ocean.context.ocean import ocean


@ocean.on_resync()
async def on_resync(kind: str) -> list[dict[Any, Any]]:
    # Get all data from the source system
    # Return raw data to run manipulation over
    return []


# Optional
@ocean.on_start()
async def on_start() -> None:
    # Something to do when the integration starts
    print("Starting integration")



from port_ocean.core.handlers.manipulation.base import BaseManipulation
from port_ocean.core.handlers.port_app_config.base import BasePortAppConfig
from port_ocean.core.handlers.transport.base import BaseTransport


__all__ = [
    "BaseManipulation",
    "BasePortAppConfig",
    "BaseTransport",
]
